
#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>

/* compile with:
 * cc -o E -Wall EXT-PASM.c EMATH.c -g -ldl
 *
 * ./E
 */

typedef int opcode_t;
static void **ops_addr;
typedef struct op_lib_t {
    int nops;
    void **func_table;
} op_lib_t;

typedef op_lib_t *(*oplib_init_f)(int init);
extern  op_lib_t *math_ops_init(int init);

struct op_libs {
    char *name;
    oplib_init_f init;
    char *initname;
    char *sofile;
    void *handle;
} op_libs[3];

static void
init_static_libs(void)
{
    op_libs[0].name = "math";
    op_libs[0].init = math_ops_init;

    op_libs[1].name = "math2";
    op_libs[1].initname = "math2_ops_init";
    op_libs[1].sofile = "EMATH2_ops.so";
}

static char *const_table[] = {	/* actually only string constants */
    "math",
    "math2"
};

static void add_lib(char *s);
opcode_t *
cg_core(opcode_t *cur_opcode)
{
    static void *lops_addr[] = {
	&&PC_0,
	&&PC_1,
	&&PC_2,
	&&PC_3,
	&&PC_4,
	NULL
    };
    if (!ops_addr)
	ops_addr = lops_addr;

    goto *ops_addr[*cur_opcode];

PC_0: /* Parrot_end */
    {
	printf("end\n");
	return (0);;
    }

PC_1: /* Parrot_noop */
    {
	printf("noop\n");
	goto *ops_addr[*(cur_opcode += 1)];
    }

PC_2: /* Parrot_oplib_sc */
    {
	printf("oplib: %s\n", const_table[cur_opcode[1]]);
	add_lib(const_table[cur_opcode[1]]);
	goto *ops_addr[*(cur_opcode += 2)];
    }

PC_3: /* Parrot_print_ic */
    {
	printf("op 3: %d\n", cur_opcode[1]);
	goto *ops_addr[*(cur_opcode += 2)];
    }

PC_4: /* Parrot_print_ic */
    {
	printf("op 4: %d\n", cur_opcode[1]);
	goto *ops_addr[*(cur_opcode += 2)];
    }
}

static struct op_lib_t core_op_lib = {
    5,
    NULL
};

struct op_lib_t * core_ops_init(int i)
{
    core_op_lib.func_table = ops_addr;
    return &core_op_lib;
}

static void add_lib(char *s)
{
    size_t i;
    int found = -1;
    op_lib_t *lib;
    oplib_init_f init_func;
    static int allocated;	/* interpreter->allocated */
    for (i = 0; i < sizeof(op_libs)/sizeof(op_libs[0]); i++) {
	if (!op_libs[i].name)
	    break;
	if (!strcmp(op_libs[i].name, s)) {
	    found = i;
	    /* TODO add check against duplicate loading */
	    if (!op_libs[i].sofile) {
		printf("*** found static lib '%s'\n", s);
		init_func = op_libs[i].init;
	    }
	    else {
		char *file = op_libs[i].sofile;
		printf("*** found dynamic lib '%s'\n", s);
		op_libs[i].handle = dlopen(file, RTLD_NOW);
		if (!op_libs[i].handle) {
		    printf("Couldn't open '%s'\n", file);
		    exit(1);
		}
		op_libs[i].init = init_func =
		    (oplib_init_f)dlsym(op_libs[i].handle,
					       op_libs[i].initname);
	    }
	    break;
	}
    }
    if (found >= 0) {
	size_t n;
	void ** new_addr;
	lib = init_func(1);
	/* if ops_addr is allocated, realloc it, else alloc */
	n = core_op_lib.nops + lib->nops;
	if (!allocated) {
	    new_addr = malloc(n * sizeof(void *));
	    for (i = 0; i < core_op_lib.nops; i++)
		new_addr[i] = ops_addr[i];
	}
	else
	    new_addr = realloc(ops_addr, n * sizeof(void *));
	for (i = core_op_lib.nops; i < n; i++)
	    new_addr[i] = lib->func_table[i - core_op_lib.nops];
	allocated = 1;
	core_op_lib.func_table = ops_addr = new_addr;
	/* tell all libs, where the jump table is */
	for (i = 0; i < sizeof(op_libs)/sizeof(op_libs[0]); i++) {
	    init_func = op_libs[i].init;
	    if (init_func)
		init_func((int)new_addr);
	    else
		break;
	}
	core_op_lib.nops = n;
    }
}

int main(int argc, char *argv[])
{
    opcode_t prog1[] = {3,10, 4,11, 1, 0};
    opcode_t prog2[] = {3,20, 4,21, 2,0, 5,10, 3,30, 6,20, 0};
    opcode_t prog3[] = {3,20, 4,21, 2,0, 2,1, 5,10, 7,10, 3,30, 8,20, 0};

    init_static_libs();
    //cg_core(prog1);
    //cg_core(prog2);
    cg_core(prog3);
    return 0;
}
